# Generated by Django 4.1.3 on 2023-01-02 02:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("formkit_ninja", "0001_initial"),
    ]

    operations = [
        # migrations.RemoveField(
        #     model_name="formkitschemanode",
        #     name="children",
        # ),
        migrations.RemoveField(
            model_name="formkitschemanode",
            name="else_condition",
        ),
        migrations.RemoveField(
            model_name="formkitschemanode",
            name="html_id",
        ),
        migrations.RemoveField(
            model_name="formkitschemanode",
            name="if_condition",
        ),
        migrations.RemoveField(
            model_name="formkitschemanode",
            name="key",
        ),
        migrations.RemoveField(
            model_name="formkitschemanode",
            name="name",
        ),
        migrations.RemoveField(
            model_name="formkitschemanode",
            name="node_value",
        ),
        migrations.RemoveField(
            model_name="formkitschemanode",
            name="placeholder",
        ),
        migrations.RemoveField(
            model_name="formkitschemanode",
            name="then_condition",
        ),
        migrations.AddField(
            model_name="formkitschemanode",
            name="values",
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="formkitschemanode",
            name="node_type",
            field=models.CharField(
                blank=True,
                choices=[
                    ("$cmp", "Component"),
                    ("text", "Text"),
                    ("condition", "Condition"),
                    ("$formkit", "FormKit"),
                    ("$el", "Element"),
                ],
                max_length=256,
            ),
        ),
    ]



# Generated by Django 4.1.3 on 2023-01-03 13:35

from django.db import migrations

import formkit_ninja.models


class Migration(migrations.Migration):

    dependencies = [
        ("formkit_ninja", "0006_formkitschemanode_placeholder"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="formcomponents",
            options={"ordering": ("schem", "order")},
        ),
        migrations.AddField(
            model_name="formkitschemanode",
            name="new_values",
            field=formkit_ninja.models.PydanticBaseModelField(blank=True, null=True),
        ),
    ]



# Generated by Django 4.1.3 on 2023-01-03 06:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("formkit_ninja", "0004_alter_formkitschemanode_group_nodechildren_and_more"),
    ]

    operations = [
        migrations.DeleteModel(
            name="SchemaCondition",
        ),
        migrations.AlterField(
            model_name="formkitschemanode",
            name="node_type",
            field=models.CharField(
                blank=True,
                choices=[
                    ("$cmp", "Component"),
                    ("text", "Text"),
                    ("condition", "Condition"),
                    ("$formkit", "FormKit"),
                    ("$el", "Element"),
                    ("raw", "Raw JSON"),
                ],
                max_length=256,
            ),
        ),
    ]



from __future__ import annotations

import logging
import warnings
from typing import Any, Optional

from django import forms
from django.conf import settings
from django.contrib import admin
from django.db import models as dj_models
from django.http import HttpRequest
from ordered_model.admin import OrderedInlineModelAdminMixin, OrderedModelAdmin, OrderedTabularInline

from formkit_ninja import models
from formkit_ninja.fields import TranslatedField
from formkit_ninja.formkit_schema import FORMKIT_TYPE

logger = logging.getLogger(__name__)


class ItemAdmin(OrderedModelAdmin):
    list_display = ("name", "move_up_down_links")


def translated_fields(form: forms.BaseForm | TransModelForm) -> tuple[str]:
    """
    List the names of fields which are TranslatedField types
    """

    def check_skip_translation_fields(model_) -> None:
        """
        A model can express "Don't translate" fields
        by means of a set `_skip_translations` on the
        form instance.
        Expect this field to be present and to be a subclass of TranslatedField
        """
        for field_name in list(getattr(form, "_skip_translations", {})):
            assert isinstance(model_._meta.get_field(field_name), TranslatedField)

    def get_model() -> dj_models.Model:
        if hasattr(form, "model"):
            return form.model
        elif hasattr(form, "Meta"):
            return form.Meta.model
        else:
            raise TypeError()

    model_ = get_model()
    check_skip_translation_fields(model_)
    fields_ = model_._meta.fields

    return tuple(
        (
            field.name
            for field in fields_
            if isinstance(field, TranslatedField) and field.name not in getattr(form, "_skip_translations", {})
        )
    )


def enlangished_class(base_form: forms.BaseForm) -> forms.BaseForm:
    """
    Enhances a 'Form' with additional fields based on settings languages
    """
    fields = translated_fields(base_form)
    label = "json_languages_{prefix}".format(prefix="_".join(fields))
    language_fields = {}
    for field in fields:
        for langcode, langname in settings.LANGUAGES:
            fieldname = "%s_%s" % (field, langcode)
            language_fields[fieldname] = forms.CharField()

    return type(label, (base_form,), language_fields)


class TransModelForm(forms.ModelForm):
    """
    Adds additional text fields to the admin where a model uses the "TranslatedField"
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for translated_field in translated_fields(self):

            # Find the initial value of the translated fields
            if kwargs.get("instance"):
                values = getattr(kwargs["instance"], translated_field) or {}
            else:
                values = {}

            for (langcode, langname) in settings.LANGUAGES:
                # Field names should be compatible with
                # JSON format ie {field}_{languuge}
                fieldname = "%s_%s" % (translated_field, langcode)
                self.fields[fieldname] = forms.CharField(
                    label=f"{translated_field} in {langname}",
                    required=False,
                    initial=values.get(langcode, ""),
                )

    def save(self, commit=True):
        for translated_field in translated_fields(self):
            translated = {}
            for (langcode, langname) in settings.LANGUAGES:
                fieldname = "%s_%s" % (translated_field, langcode)
                value = self.cleaned_data[fieldname]
                if value:
                    translated[langcode] = value
            setattr(self.instance, translated_field, translated)
        return super().save(commit=commit)


class JsonDecoratedFormBase(TransModelForm):
    """
    Adds additional fields to the admin where a model has a JSON field
    and some appropriate (tbc?) field parameters
    """

    # extra = forms.CharField(label="Extra", max_length=128, required=False)
    # hello_world = forms.CharField(widget=forms.NumberInput, required=False)

    # key is the name of a `models.JSONField` on the model
    # value is a list of fields to get/set in that JSON field
    _json_fields: dict[str, tuple[str]] = {"my_json_field": ("formkit", "description", "name", "key", "html_id")}

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        for field, keys in self._json_fields.items():
            # Extract the dict of JSON values from the model instance if supplied
            values = getattr(kwargs["instance"], field) or {} if kwargs.get("instance") else {}
            for key in keys:
                # The value, extracted from the JSON value in the database
                field_value = values.get(key, None)
                # The initial value of the admin form is set to the value of the JSON attrib
                if key in self.fields:
                    self.fields.get(key).initial = field_value
                else:
                    warnings.warn(f"Unassociated JSON field found: {key}")
        return

    def save(self, commit=True):
        """
        Updates the JSON field(s) from the fields specified in the `_json_fields` dict
        """

        for field, keys in self._json_fields.items():
            # Populate a JSON field in a model named "form"
            # from a set of standard form elements
            data = getattr(self.instance, field, {})
            if not isinstance(data, dict):
                # This is a bit unexpected
                raise ValueError("Expected a JSON dict object here")
            logger.debug(data)
            data.update({key: self.cleaned_data[key] for key in keys})
            logger.debug(data)
            setattr(self.instance, field, data)
        return super().save(commit=commit)


class NewFormKitForm(forms.ModelForm):
    class Meta:
        model = models.FormKitSchemaNode
        fields = ("node_type", "description")


class OptionForm(TransModelForm):
    class Meta:
        model = models.Option
        exclude = ("label",)


class FormComponentsForm(TransModelForm):
    class Meta:
        model = models.FormComponents
        exclude = ()


class FormKitSchemaNodeOptionsInline(OrderedTabularInline):
    model = models.Option
    form = enlangished_class(OptionForm)
    readonly_fields = (
        "order",
        "move_up_down_links",
    )
    ordering = ("order",)
    extra = 1


class FormKitSchemaComponentInline(OrderedTabularInline):
    model = models.FormComponents
    # form = FormComponentsForm
    readonly_fields = (
        "order",
        "move_up_down_links",
        "node",
    )
    ordering = ("order",)
    extra = 1


class FormKitNodeGroupForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = ("description",)

    _json_fields = {
        "node": (
            "formkit",
            "if_condition",
        )
    }

    formkit = forms.ChoiceField(required=False, choices=models.FormKitSchemaNode.FORMKIT_CHOICES, disabled=True)
    if_condition = forms.CharField(
        widget=forms.TextInput,
        required=False,
    )


class FormKitNodeForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = ("description",)

    _json_fields = {"node": ("formkit", "description", "name", "key", "html_id", "if_condition")}

    formkit = forms.ChoiceField(required=False, choices=models.FormKitSchemaNode.FORMKIT_CHOICES)
    name = forms.CharField(
        required=False,
    )
    if_condition = forms.CharField(
        widget=forms.TextInput,
        required=False,
    )
    key = forms.CharField(
        required=False,
    )
    html_id = forms.CharField(
        required=False,
        help_text="Use this ID if adding conditions to other fields (hint: $get(my_field).value === 8)",
    )


class FormKitTextNode(TransModelForm):
    class Meta:
        model = models.FormKitSchemaNode
        fields = ("description",)


class FormKitElementForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = ("description",)

    _skip_translations = {"label", "placeholder"}
    _json_fields = {"node": ("el", "name", "if_condition", "classes")}

    el = forms.ChoiceField(required=False, choices=models.FormKitSchemaNode.ELEMENT_TYPE_CHOICES)
    name = forms.CharField(
        required=False,
    )
    classes = forms.CharField(
        required=False,
    )
    if_condition = forms.CharField(
        widget=forms.TextInput,
        required=False,
    )


class FormKitConditionForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        # fields = '__all__'
        fields = ("description",)

    _json_fields = {"node": ("if_condition", "then_condition", "else_condition")}

    if_condition = forms.CharField(
        widget=forms.TextInput,
        required=False,
    )
    then_condition = forms.CharField(
        max_length=256,
        required=False,
    )
    else_condition = forms.CharField(
        max_length=256,
        required=False,
    )


class FormKitComponentForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = ("description",)

    _json_fields = {"node": ("if_condition", "then_condition", "else_condition")}


class MembershipInline(OrderedTabularInline):
    model = models.Membership
    fk_name = "group"
    extra = 0
    readonly_fields = (
        "order",
        "move_up_down_links",
    )
    ordering = ("order",)


class MembershipComponentInline(admin.StackedInline):
    model = models.Membership
    fk_name = "member"
    extra = 0


class NodeChildrenInline(OrderedTabularInline):
    """
    Nested HTML elements
    """

    model = models.NodeChildren
    readonly_fields = (
        "order",
        "move_up_down_links",
    )
    ordering = ("order",)
    fk_name = "parent"
    extra = 0


class FormKitSchemaForm(TransModelForm):
    class Meta:
        model = models.FormKitSchema
        exclude = ("name",)


@admin.register(models.FormKitSchemaNode)
class FormKitSchemaNodeAdmin(OrderedInlineModelAdminMixin, admin.ModelAdmin):
    def get_inlines(self, request, obj: models.FormKitSchemaNode | None):

        if not obj:
            return []

        formkit_node_type = (obj.node or {}).get("formkit", None)

        if formkit_node_type == "group":
            return [
                MembershipInline,
            ]
        elif formkit_node_type in {"radio", "select"}:
            return [
                MembershipComponentInline,
                FormKitSchemaNodeOptionsInline,
            ]
        elif formkit_node_type:
            return [
                MembershipComponentInline,
            ]

        if obj.node_type == "$el":
            return [
                NodeChildrenInline,
            ]

        else:
            return []

    # Note that although overridden these are necessary
    inlines = [FormKitSchemaNodeOptionsInline, NodeChildrenInline, MembershipInline]

    def get_fieldsets(
        self, request: HttpRequest, obj: Optional[models.FormKitSchemaNode] = ...
    ) -> list[tuple[Optional[str], dict[str, Any]]]:

        return super().get_fieldsets(request, obj)

    def get_form(
        self,
        request: Any,
        obj: models.FormKitSchemaNode | None,
        change: None = None,
        **kwargs,
    ):
        if not obj:
            return NewFormKitForm

        try:
            node_type: FORMKIT_TYPE = obj.node.get("node_type")
        except (AttributeError, KeyError) as E:
            warnings.warn("Expected a 'Node' with a 'NodeType' in the admin form")
            warnings.warn(f"{E}")

        if node_type == "condition":
            return FormKitConditionForm

        elif node_type == "formkit":
            formkit_node_type = (obj.node or {}).get("formkit", None)
            if formkit_node_type == "group":
                return enlangished_class(FormKitNodeGroupForm)
            return enlangished_class(FormKitNodeForm)

        elif node_type == "el":
            return enlangished_class(FormKitElementForm)

        elif node_type == "text":
            return enlangished_class(FormKitTextNode)

        elif node_type == "component":
            return enlangished_class(FormKitComponentForm)

        else:
            return super().get_form(request, obj, change, **kwargs)


@admin.register(models.Option)
class OptionAdmin(OrderedModelAdmin):
    form = enlangished_class(OptionForm)
    list_display = ("label", "field", "move_up_down_links")


@admin.register(models.FormKitSchema)
class FormKitSchemaAdmin(OrderedInlineModelAdminMixin, admin.ModelAdmin):
    form = enlangished_class(FormKitSchemaForm)

    def get_inlines(self, request, obj: models.FormKitSchema | None):

        if not obj:
            return []
        return [
            FormKitSchemaComponentInline,
        ]

    inlines = [
        FormKitSchemaComponentInline,
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['formkit_ninja',
 'formkit_ninja.management.commands',
 'formkit_ninja.migrations']

package_data = \
{'': ['*']}

install_requires = \
['Django>=4.1.5,<5.0.0',
 'django-ninja>=0.20.0,<0.21.0',
 'django-ordered-model>=3.6,<4.0',
 'pydantic<2']

setup_kwargs = {
    'name': 'formkit-ninja',
    'version': '0.1.6a3',
    'description': 'A Django-Ninja backend to specify FormKit schemas',
    'long_description': '# Formkit-Ninja\n\nA Django-Ninja framework for FormKit schemas and form submissions\n\n## Why\n\nFormKit out of the box has awesome schema support - this lets us integrate FormKit instances as Django models\n\n- Upload / edit / download basic FormKit schemas\n- Translated "option" values from the Django admin\n- Reorder "options" and schema nodes\n- List and Fetch schemas for different form types\n\n## Use\n\nTo use, `pip install formkit-ninja` and add the following to settings `INSTALLED_APPS`:\n\n```py\nINSTALLED_APPS = [\n    ...\n    "formkit_ninja",\n    "ordered_model",\n    "ninja",\n    ...\n]\n```\n\n## Test\n\nPull the repo:\n\n`gh repo clone catalpainternational/formkit-ninja`\n`cd formkit-ninja`\n`poetry install`\n`poetry run pytest`\n\n## Lint\n\n`poetry run black --check .`\n`poetry run isort --check .`\n`poetry run flake8 .`\n',
    'author': 'Josh Brooks',
    'author_email': 'josh@catalpa.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/catalpainternational/formkit-ninja',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import logging
from html.parser import HTMLParser
from typing import Annotated, Any, ForwardRef, List, Literal, Type, TypedDict, TypeVar, Union

from pydantic import BaseModel, Field

"""
This is a port of selected parts of the FormKit schema
to Pydantic models.
"""


logger = logging.getLogger(__name__)

HtmlAttrs = dict[str, str | dict[str, str]]


Node = ForwardRef("Node")


class FormKitSchemaCondition(BaseModel):
    node_type: Literal["condition"]
    if_condition: str = Field(..., alias="if")
    then_condition: Node | List[Node] = Field(..., alias="then")
    else_condition: Node | List[Node] | None = Field(None, alias="else")


class FormKitSchemaConditionNoCircularRefs(BaseModel):
    """
    This class is defined in order to break circular references.
    Circular references (ie the `Node` in the then_condition) cause
    django_ninja to crash the server hard
    """

    node_type: Literal["condition"]
    if_condition: str = Field(..., alias="if")
    then_condition: str | list[str] = Field(..., alias="then")
    else_condition: str | list[str] | None = Field(None, alias="else")


class FormKitSchemaMeta(BaseModel):
    __root__: dict[str, str | float | int | bool | None]


class FormKitTypeDefinition(BaseModel):
    ...


class FormKitContextShape(BaseModel):
    type: Literal["input", "list", "group"]
    value: Any
    _value: Any


class FormKitListValue(BaseModel):
    __root__: str | list[str] | list[dict[str, str]]


class FormKitListStatement(BaseModel):
    """
    A full loop statement in tuple syntax. Can be read like "foreach value, key? in list"
    A 2 or 2 element tuple of value, key, and list or value, list
    """

    __root__: tuple[str, float | int | str, list[FormKitListValue]]


class FormKitSchemaAttributesCondition(BaseModel):
    if_: str = Field(alias="if")
    then_: FormKitAttributeValue = Field(alias="then")
    else_: FormKitAttributeValue | None = Field(alias="else")

    class Config:
        allow_population_by_field_name = True


class FormKitAttributeValue(BaseModel):
    """
    The possible value types of attributes (in the schema)
    """

    __root__: str | int | float | bool | None | FormKitSchemaAttributes | FormKitSchemaAttributesCondition


# ForwardRef allows for self referencing (recursive) models
FormKitSchemaAttributes = ForwardRef("FormKitSchemaAttributes")


class FormKitSchemaAttributes(BaseModel):
    __root__: dict[
        str,
        FormKitAttributeValue
        # | FormKitSchemaAttributes  # <-- Causes trouble for Ninja
        | FormKitSchemaAttributesCondition,
    ]


class FormKitSchemaProps(BaseModel):
    """
    Properties available in all schema nodes.
    """

    # "ForwardRefs" do not work well with django-ninja.
    # This would ideally be:
    # children: str | list[FormKitSchemaProps] | FormKitSchemaCondition | None = Field(
    #     default_factory=list
    # )
    children: str | list[FormKitSchemaProps] | FormKitSchemaConditionNoCircularRefs | None = Field()
    key: str | None
    if_condition: str | None = Field(alias="if")
    for_loop: FormKitListStatement | None = Field(alias="for")
    bind: str | None
    meta: FormKitSchemaMeta | None

    # These are not formal parts of spec, but
    # are attributes defined in ts as Record<string, any>
    # id: str | uuid.UUID | None = Field(None)
    html_id: str = Field(None, alias="id")
    name: str | None = Field(None)
    label: str | None = Field(None)
    help: str | None = Field(None)
    validation: str | None = Field(None)
    validationLabel: str | None = Field(None, alias="validation-label")
    validationVisibility: str | None = Field(None, alias="validation-visibility")
    placeholder: str | None = Field(None)
    value: str | None = Field(None)
    prefixIcon: str | None = Field(None, alias="prefix-icon")

    class Config:
        allow_population_by_field_name = True


class TextNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["text"] = "text"
    dollar_formkit: str = Field(default="text", alias="$formkit")
    text: str | None


class DateNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["date"] = "date"
    dollar_formkit: str = Field(default="date", alias="$formkit")


class CheckBoxNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["checkbox"] = "checkbox"
    dollar_formkit: str = Field(default="checkbox", alias="$formkit")


class NumberNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["number"] = "number"
    dollar_formkit: str = Field(default="number", alias="$formkit")
    text: str | None


class PasswordNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["password"] = "password"
    dollar_formkit: str = Field(default="password", alias="$formkit")
    name: str | None


class RadioNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["radio"] = "radio"
    dollar_formkit: str = Field(default="radio", alias="$formkit")
    name: str | None
    options: list[dict[str, Any]] | list[str] | dict[str, str] | None = Field(None)


class SelectNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["select"] = "select"
    dollar_formkit: str = Field(default="select", alias="$formkit")
    options: list[dict[str, Any]] | list[str] | dict[str, str] | None = Field(None)


class EmailNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["email"] = "email"
    dollar_formkit: str = Field(default="email", alias="$formkit")


class GroupNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["group"] = "group"
    dollar_formkit: str = Field(default="group", alias="$formkit")
    text: str | None


FormKitSchemaFormKit = Annotated[
    Union[TextNode, CheckBoxNode, PasswordNode, SelectNode, EmailNode, NumberNode, RadioNode, GroupNode, DateNode],
    Field(discriminator="formkit"),
]


class FormKitSchemaDOMNode(FormKitSchemaProps):
    """
    HTML elements are defined using the $el property.
    You can use $el to render any HTML element.
    Attributes can be added with the attrs property,
    and content is assigned with the children property
    """

    node_type: Literal["element"]
    el: str = Field(alias="$el")
    attrs: FormKitSchemaAttributes | None

    class Config:
        allow_population_by_field_name = True


class FormKitSchemaComponent(FormKitSchemaProps):
    """
    Components can be defined with the $cmp property
    The $cmp property should be a string that references
    a globally defined component or a component passed
    into FormKitSchema with the library prop.
    """

    node_type: Literal["component"]

    cmp: str = Field(
        ...,
        alias="$cmp",
        description="The $cmp property should be a string that references a globally defined component or a component passed into FormKitSchema with the library prop.",  # noqa: E501
    )
    props: dict[str, str | Any] | None

    class Config:
        allow_population_by_field_name = True


# This necessary to properly "populate" some more complicated models
FormKitSchemaAttributesCondition.update_forward_refs()
FormKitAttributeValue.update_forward_refs()
FormKitSchemaDOMNode.update_forward_refs()
FormKitSchemaCondition.update_forward_refs()
FormKitSchemaComponent.update_forward_refs()


class FormKitTagParser(HTMLParser):
    """
    Reverse an HTML example to schema
    This is for lazy copy-pasting from the formkit website :)
    """

    def __init__(self, html_content, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data: str | None = None

        self.current_tag: FormKitSchemaFormKit | None = None
        self.tags: list[FormKitSchemaFormKit] = []
        self.parents: list[FormKitSchemaFormKit] = []
        self.feed(html_content)

    def handle_starttag(self, tag, attrs):
        """
        Read anything that's a "formtag" type
        """
        if tag != "formkit":
            return
        props = dict(attrs)
        props["formkit"] = props.pop("type")

        tag = FormKitSchemaFormKit(**props)
        self.current_tag = tag

        if self.parents:
            self.parents[-1].children.append(tag)
        else:
            self.tags.append(tag)
            self.parents.append(tag)

    def handle_endtag(self, tag: str) -> None:
        if tag != "formkit":
            return
        if self.parents:
            self.parents.pop()

    def handle_data(self, data):
        if self.current_tag and data.strip():
            self.current_tag.children.append(data.strip())
            # Ensure that children is included even when "exclude_unset" is True
            # since we populated this after the initial tag build
            self.current_tag.__fields_set__.add("children")


FormKitSchemaDOMNode.update_forward_refs()

Model = TypeVar("Model", bound="BaseModel")
StrBytes = str | bytes

Node = Annotated[
    Union[
        FormKitSchemaFormKit,
        FormKitSchemaDOMNode,
        FormKitSchemaComponent,
        FormKitSchemaCondition,
    ],
    Field(discriminator="node_type"),
]

NODE_TYPE = Literal["condition", "formkit", "element", "component"]
FORMKIT_TYPE = Literal[
    "select", "checkbox", "number", "group", "list", "password", "button", "select", "radio", "form", "date"
]


class Discriminators(TypedDict, total=False):
    node_type: NODE_TYPE
    formkit: FORMKIT_TYPE
    dollar_formkit: FORMKIT_TYPE


def get_node_type(obj: dict) -> Discriminators:
    """
    Pydantic requires nodes to be "differentiated" by a field value
    when used in a Union type situation.
    This function should return the 'node_type' values and if present 'Formkit' value
    which corresponds to the object being inspected.
    """
    fields: Discriminators = {}
    if "__root__" in obj:
        obj = obj["__root__"]
    for key, return_value in (
        # -- Loading from "raw json"
        ("$el", "element"),
        ("$formkit", "formkit"),
        ("dollar_formkit", "formkit"),
        ("$cmp", "component"),
        # -- Already in the database
        ("el", "element"),
        ("formkit", "formkit"),
        ("cmp", "component"),
        ("if", "condition"),
    ):
        if key in obj:
            fields["node_type"] = return_value
            # "formkit" and "$formkit" are aliases
            # Because we use these in a disriminated union, we can't alias
            # So for consistency: we always return both
            if key == "$formkit":
                fields["formkit"] = obj["$formkit"]
                fields["dollar_formkit"] = obj["$formkit"]
            elif key == "dollar_formkit":
                fields["formkit"] = obj["dollar_formkit"]
                fields["dollar_formkit"] = obj["dollar_formkit"]

            return fields
    raise KeyError("Could not determine node type")


class FormKitNode(BaseModel):
    __root__: Node

    @classmethod
    def parse_obj(cls: Type["Model"], obj: Any) -> "Model":
        if "id" in obj:
            obj["html_id"] = obj.pop("id")
        return super().parse_obj({**get_node_type(obj), **obj})


class FormKitSchema(BaseModel):
    __root__: list[Node]

    @classmethod
    def parse_obj(cls: Type["Model"], obj: Any) -> "Model":
        return super().parse_obj([{**get_node_type(obj), **obj} for obj in obj])


FormKitSchema.update_forward_refs()
FormKitSchemaCondition.update_forward_refs()
PasswordNode.update_forward_refs()
